@@ -60,6 +60,10 @@ module Sequel
       def connect(server)
         setup_connection(::DataObjects::Connection.new(uri(server_opts(server))))
       end
+
+      def disconnect_connection(conn)
+        conn.dispose
+      end
       
       # Execute the given SQL.  If a block is given, the DataObjects::Reader
       # created is yielded to it. A block should not be provided unless a
@@ -126,11 +130,6 @@ module Sequel
         [::DataObjects::Error]
       end
 
-      # Close the given database connection.
-      def disconnect_connection(c)
-        c.close
-      end
-
       # Recognize DataObjects::ConnectionError instances as disconnect errors.
       def disconnect_error?(e, opts)
         super || (e.is_a?(::DataObjects::Error) && (e.is_a?(::DataObjects::ConnectionError) || e.message =~ DISCONNECT_ERROR_RE))
