@@ -5,25 +5,31 @@ module Sequel
     # Database and Dataset instance methods for JTDS specific
     # support via JDBC.
     module JTDS
-      # Database instance methods for JTDS databases accessed via JDBC.
       module DatabaseMethods
         include Sequel::JDBC::MSSQL::DatabaseMethods
-        
-        # Return instance of Sequel::JDBC::JTDS::Dataset with the given opts.
-        def dataset(opts=nil)
-          Sequel::JDBC::JTDS::Dataset.new(self, opts)
+
+        private
+
+        # Handle nil values by using setNull with the correct parameter type.
+        def set_ps_arg_nil(cps, i)
+          cps.setNull(i, cps.getParameterMetaData.getParameterType(i))
         end
       end
-      
+
       # Dataset class for JTDS datasets accessed via JDBC.
       class Dataset < JDBC::Dataset
         include Sequel::MSSQL::DatasetMethods
 
+        class ::Sequel::JDBC::Dataset::TYPE_TRANSLATOR
+          def jtds_clob(v) v.getSubString(1, v.length) end
+        end
+
+        JTDS_CLOB_METHOD = TYPE_TRANSLATOR_INSTANCE.method(:jtds_clob)
+      
         # Handle CLOB types retrieved via JTDS.
-        def convert_type(v)
-          case v
-          when Java::NetSourceforgeJtdsJdbc::ClobImpl
-            convert_type(v.getSubString(1, v.length))
+        def convert_type_proc(v)
+          if v.is_a?(Java::NetSourceforgeJtdsJdbc::ClobImpl)
+            JTDS_CLOB_METHOD
           else
             super
           end
