@@ -1,3 +1,5 @@
+Sequel.require 'adapters/utils/pg_types'
+
 module Sequel
   # Top level module for holding all PostgreSQL-related modules and classes
   # for Sequel.  There are a few module level accessors that are added via
@@ -94,6 +96,8 @@ module Sequel
       RE_CURRVAL_ERROR = /currval of sequence "(.*)" is not yet defined in this session|relation "(.*)" does not exist/.freeze
       SYSTEM_TABLE_REGEXP = /^pg|sql/.freeze
       FOREIGN_KEY_LIST_ON_DELETE_MAP = {'a'.freeze=>:no_action, 'r'.freeze=>:restrict, 'c'.freeze=>:cascade, 'n'.freeze=>:set_null, 'd'.freeze=>:set_default}.freeze
+      POSTGRES_DEFAULT_RE = /\A(?:B?('.*')::[^']+|\((-?\d+(?:\.\d+)?)\))\z/
+      UNLOGGED = 'UNLOGGED '.freeze
 
       # SQL fragment for custom sequences (ones not created by serial primary key),
       # Returning the schema and literal form of the sequence name, by parsing
@@ -134,7 +138,7 @@ module Sequel
       SELECT_SERIAL_SEQUENCE_SQL = (<<-end_sql
         SELECT  name.nspname AS "schema", seq.relname AS "sequence"
         FROM pg_class seq, pg_attribute attr, pg_depend dep,
-          pg_namespace name, pg_constraint cons
+          pg_namespace name, pg_constraint cons, pg_class t
         WHERE seq.oid = dep.objid
           AND seq.relnamespace  = name.oid
           AND seq.relkind = 'S'
@@ -142,10 +146,18 @@ module Sequel
           AND attr.attnum = dep.refobjsubid
           AND attr.attrelid = cons.conrelid
           AND attr.attnum = cons.conkey[1]
+          AND attr.attrelid = t.oid
           AND cons.contype = 'p'
       end_sql
       ).strip.gsub(/\s+/, ' ').freeze
 
+      # The Sequel extensions that require reseting of the conversion procs.
+      RESET_PROCS_EXTENSIONS = [:pg_array, :pg_hstore, :pg_inet, :pg_interval, :pg_json, :pg_range].freeze
+
+      # A hash of conversion procs, keyed by type integer (oid) and
+      # having callable values for the conversion proc for that type.
+      attr_reader :conversion_procs
+
       # Commit an existing prepared transaction with the given transaction
       # identifier string.
       def commit_prepared_transaction(transaction_id)
@@ -212,6 +224,16 @@ module Sequel
         :postgres
       end
 
+      # Use PostgreSQL's DO syntax to execute an anonymous code block.  The code should
+      # be the literal code string to use in the underlying procedural language.  Options:
+      #
+      # :language :: The procedural language the code is written in.  The PostgreSQL
+      #              default is plpgsql.  Can be specified as a string or a symbol.
+      def do(code, opts={})
+        language = opts[:language]
+        run "DO #{"LANGUAGE #{literal(language.to_s)} " if language}#{literal(code)}"
+      end
+
       # Drops the function from the database. Arguments:
       # * name : name of the function to drop
       # * opts : options hash:
@@ -250,6 +272,17 @@ module Sequel
         self << drop_trigger_sql(table, name, opts)
       end
 
+      # If any of the extensions that require reseting the conversion procs
+      # is loaded, reset them.  This is done here so that if you load
+      # multiple pg_* extensions in the same call, the conversion procs are
+      # only reset once instead of once for every extension.
+      def extension(*exts)
+        super
+        unless (RESET_PROCS_EXTENSIONS & exts).empty?
+          reset_conversion_procs
+        end
+      end
+
       # Return full foreign key information using the pg system tables, including
       # :name, :on_delete, :on_update, and :deferrable entries in the hashes.
       def foreign_key_list(table, opts={})
@@ -320,16 +353,17 @@ module Sequel
           join(:pg_index___ind, :indrelid=>:oid, im.call(table)=>:relname).
           join(:pg_class___indc, :oid=>:indexrelid).
           join(:pg_attribute___att, :attrelid=>:tab__oid, :attnum=>attnums).
+          left_join(:pg_constraint___con, :conname=>:indc__relname).
           filter(:indc__relkind=>'i', :ind__indisprimary=>false, :indexprs=>nil, :indpred=>nil, :indisvalid=>true).
           order(:indc__relname, SQL::CaseExpression.new(range.map{|x| [SQL::Subscript.new(:ind__indkey, [x]), x]}, 32, :att__attnum)).
-          select(:indc__relname___name, :ind__indisunique___unique, :att__attname___column)
+          select(:indc__relname___name, :ind__indisunique___unique, :att__attname___column, :con__condeferrable___deferrable)
 
         ds.join!(:pg_namespace___nsp, :oid=>:tab__relnamespace, :nspname=>schema.to_s) if schema
         ds.filter!(:indisready=>true, :indcheckxmin=>false) if server_version >= 80300
 
         indexes = {}
         ds.each do |r|
-          i = indexes[m.call(r[:name])] ||= {:columns=>[], :unique=>r[:unique]}
+          i = indexes[m.call(r[:name])] ||= {:columns=>[], :unique=>r[:unique], :deferrable=>r[:deferrable]}
           i[:columns] << m.call(r[:column])
         end
         indexes
@@ -347,47 +381,56 @@ module Sequel
       # :server :: The server to which to send the NOTIFY statement, if the sharding support
       #            is being used.
       def notify(channel, opts={})
-        execute_ddl("NOTIFY #{channel}#{", #{literal(opts[:payload].to_s)}" if opts[:payload]}", opts)
+        execute_ddl("NOTIFY #{dataset.send(:table_ref, channel)}#{", #{literal(opts[:payload].to_s)}" if opts[:payload]}", opts)
       end
 
       # Return primary key for the given table.
       def primary_key(table, opts={})
         quoted_table = quote_schema_table(table)
-        @primary_keys.fetch(quoted_table) do
-          schema, table = schema_and_table(table)
-          sql = "#{SELECT_PK_SQL} AND pg_class.relname = #{literal(table)}"
-          sql << "AND pg_namespace.nspname = #{literal(schema)}" if schema
-          @primary_keys[quoted_table] = fetch(sql).single_value
-        end
+        Sequel.synchronize{return @primary_keys[quoted_table] if @primary_keys.has_key?(quoted_table)}
+        schema, table = schema_and_table(table)
+        sql = "#{SELECT_PK_SQL} AND pg_class.relname = #{literal(table)}"
+        sql << " AND pg_namespace.nspname = #{literal(schema)}" if schema
+        value = fetch(sql).single_value
+        Sequel.synchronize{@primary_keys[quoted_table] = value}
       end
 
       # Return the sequence providing the default for the primary key for the given table.
       def primary_key_sequence(table, opts={})
         quoted_table = quote_schema_table(table)
-        @primary_key_sequences.fetch(quoted_table) do
-          schema, table = schema_and_table(table)
-          table = literal(table)
-          sql = "#{SELECT_SERIAL_SEQUENCE_SQL} AND seq.relname = #{table}"
+        Sequel.synchronize{return @primary_key_sequences[quoted_table] if @primary_key_sequences.has_key?(quoted_table)}
+        schema, table = schema_and_table(table)
+        table = literal(table)
+        sql = "#{SELECT_SERIAL_SEQUENCE_SQL} AND t.relname = #{table}"
+        sql << " AND name.nspname = #{literal(schema)}" if schema
+        if pks = fetch(sql).single_record
+          value = literal(SQL::QualifiedIdentifier.new(pks[:schema], pks[:sequence]))
+          Sequel.synchronize{@primary_key_sequences[quoted_table] = value}
+        else
+          sql = "#{SELECT_CUSTOM_SEQUENCE_SQL} AND t.relname = #{table}"
           sql << " AND name.nspname = #{literal(schema)}" if schema
-          unless pks = fetch(sql).single_record
-            sql = "#{SELECT_CUSTOM_SEQUENCE_SQL} AND t.relname = #{table}"
-            sql << " AND name.nspname = #{literal(schema)}" if schema
-            pks = fetch(sql).single_record
-          end
-
-          @primary_key_sequences[quoted_table] = if pks
-            literal(SQL::QualifiedIdentifier.new(pks[:schema], LiteralString.new(pks[:sequence])))
+          if pks = fetch(sql).single_record
+            value = literal(SQL::QualifiedIdentifier.new(pks[:schema], LiteralString.new(pks[:sequence])))
+            Sequel.synchronize{@primary_key_sequences[quoted_table] = value}
           end
         end
       end
+      
+      # Reset the database's conversion procs, requires a server query if there
+      # any named types.
+      def reset_conversion_procs
+        @conversion_procs = get_conversion_procs
+      end
 
-      # Reset the primary key sequence for the given table, baseing it on the
+      # Reset the primary key sequence for the given table, basing it on the
       # maximum current value of the table's primary key.
       def reset_primary_key_sequence(table)
         return unless seq = primary_key_sequence(table)
         pk = SQL::Identifier.new(primary_key(table))
         db = self
         seq_ds = db.from(LiteralString.new(seq))
+        s, t = schema_and_table(table)
+        table = Sequel.qualify(s, t) if s
         get{setval(seq, db[table].select{coalesce(max(pk)+seq_ds.select{:increment_by}, seq_ds.select(:min_value))}, false)}
       end
 
@@ -425,6 +468,16 @@ module Sequel
         server_version >= 90100
       end
 
+      # PostgreSQL 9.0+ supports some types of deferrable constraints beyond foreign key constraints.
+      def supports_deferrable_constraints?
+        server_version >= 90000
+      end
+
+      # PostgreSQL supports deferrable foreign key constraints.
+      def supports_deferrable_foreign_key_constraints?
+        true
+      end
+
       # PostgreSQL supports DROP TABLE IF EXISTS
       def supports_drop_table_if_exists?
         true
@@ -487,7 +540,7 @@ module Sequel
       end
     
       # Handle :using option for set_column_type op, and the :validate_constraint op.
-      def alter_table_sql(table, op)
+      def alter_table_op_sql(table, op)
         case op[:op]
         when :set_column_type
           s = super
@@ -498,7 +551,7 @@ module Sequel
           end
           s
         when :validate_constraint
-          "ALTER TABLE #{quote_schema_table(table)} VALIDATE CONSTRAINT #{quote_identifier(op[:name])}"
+          "VALIDATE CONSTRAINT #{quote_identifier(op[:name])}"
         else
           super
         end
@@ -523,6 +576,14 @@ module Sequel
         end
       end
       
+      # Handle PostgreSQL specific default format.
+      def column_schema_normalize_default(default, type)
+        if m = POSTGRES_DEFAULT_RE.match(default)
+          default = m[1] || m[2]
+        end
+        super(default, type)
+      end
+
       # If the :prepare option is given and we aren't in a savepoint,
       # prepare the transaction for a two-phase commit.
       def commit_transaction(conn, opts={})
@@ -533,6 +594,12 @@ module Sequel
         end
       end
 
+      # PostgreSQL can't combine rename_column operations, and it can combine
+      # the custom validate_constraint operation.
+      def combinable_alter_table_op?(op)
+        (super || op[:op] == :validate_constraint) && op[:op] != :rename_column
+      end
+
       # The SQL queries to execute when starting a new connection.
       def connection_configuration_sqls
         sqls = []
@@ -560,6 +627,42 @@ module Sequel
         end
       end
 
+      # SQL for doing fast table insert from stdin.
+      def copy_into_sql(table, opts)
+        sql = "COPY #{literal(table)}"
+        if cols = opts[:columns]
+          sql << literal(Array(cols))
+        end
+        sql << " FROM STDIN"
+        if opts[:options] || opts[:format]
+          sql << " ("
+          sql << "FORMAT #{opts[:format]}" if opts[:format]
+          sql << "#{', ' if opts[:format]}#{opts[:options]}" if opts[:options]
+          sql << ')'
+        end
+        sql
+      end
+
+      # SQL for doing fast table output to stdout.
+      def copy_table_sql(table, opts)
+         if table.is_a?(String)
+           return table
+         else
+           if opts[:options] || opts[:format]
+             options = " ("
+             options << "FORMAT #{opts[:format]}" if opts[:format]
+             options << "#{', ' if opts[:format]}#{opts[:options]}" if opts[:options]
+             options << ')'
+           end
+           table = if table.is_a?(::Sequel::Dataset)
+             "(#{table.sql})"
+           else
+             literal(table)
+           end
+          return "COPY #{table} TO STDOUT#{options}"
+         end
+       end
+
       # SQL statement to create database function.
       def create_function_sql(name, definition, opts={})
         args = opts[:args]
@@ -591,6 +694,17 @@ module Sequel
         "CREATE SCHEMA #{quote_identifier(name)}"
       end
 
+      # DDL statement for creating a table with the given name, columns, and options
+      def create_table_prefix_sql(name, options)
+        temp_or_unlogged_sql = if options[:temp]
+         raise(Error, "can't provide both :temp and :unlogged to create_table") if options[:unlogged]
+         temporary_table_sql
+       elsif options[:unlogged]
+         UNLOGGED
+       end
+        "CREATE #{temp_or_unlogged_sql}TABLE#{' IF NOT EXISTS' if options[:if_not_exists]} #{options[:temp] ? quote_identifier(name) : quote_schema_table(name)}"
+      end
+
       # Use a PostgreSQL-specific create table generator
       def create_table_generator_class
         Postgres::CreateTableGenerator
@@ -643,6 +757,18 @@ module Sequel
         end
       end
 
+      # Return a hash with oid keys and callable values, used for converting types.
+      def get_conversion_procs
+        procs = PG_TYPES.dup
+        procs[1184] = procs[1114] = method(:to_application_timestamp)
+        unless (pgnt = PG_NAMED_TYPES).empty?
+          from(:pg_type).where(:typtype=>'b', :typname=>pgnt.keys.map{|t| t.to_s}).select_map([:oid, :typname]).each do |oid, name|
+            procs[oid.to_i] ||= pgnt[name.untaint.to_sym]
+          end
+        end
+        procs
+      end
+
       # PostgreSQL folds unquoted identifiers to lowercase, so it shouldn't need to upcase identifiers on input.
       def identifier_input_method_default
         nil
@@ -676,6 +802,14 @@ module Sequel
         "CREATE #{unique}INDEX#{' CONCURRENTLY' if index[:concurrently]} #{quote_identifier(index_name)} ON #{quote_schema_table(table_name)} #{"USING #{index_type} " if index_type}#{expr}#{filter}"
       end
 
+      # Setup datastructures shared by all postgres adapters.
+      def initialize_postgres_adapter
+        @primary_keys = {}
+        @primary_key_sequences = {}
+        @conversion_procs = PG_TYPES.dup
+        reset_conversion_procs
+      end
+
       # Backbone of the tables and views support.
       def pg_class_relname(type, opts)
         ds = metadata_dataset.from(:pg_class).filter(:relkind=>type).select(:relname).exclude(SQL::StringExpression.like(:relname, SYSTEM_TABLE_REGEXP)).server(opts[:server]).join(:pg_namespace, :oid=>:relnamespace)
@@ -694,8 +828,10 @@ module Sequel
       # changed.
       def remove_cached_schema(table)
         tab = quote_schema_table(table)
-        @primary_keys.delete(tab)
-        @primary_key_sequences.delete(tab)
+        Sequel.synchronize do
+          @primary_keys.delete(tab)
+          @primary_key_sequences.delete(tab)
+        end
         super
       end
 
@@ -711,11 +847,22 @@ module Sequel
         super && schema[:default] =~ /\Anextval/io
       end
 
+      # Recognize PostgreSQL interval type.
+      def schema_column_type(db_type)
+        case db_type
+        when /\Ainterval\z/io
+          :interval
+        else
+          super
+        end
+      end
+
       # The dataset used for parsing table schemas, using the pg_* system catalogs.
       def schema_parse_table(table_name, opts)
         m = output_identifier_meth(opts[:dataset])
         m2 = input_identifier_meth(opts[:dataset])
         ds = metadata_dataset.select(:pg_attribute__attname___name,
+            SQL::Cast.new(:pg_attribute__atttypid, :integer).as(:oid),
             SQL::Function.new(:format_type, :pg_type__oid, :pg_attribute__atttypmod).as(:db_type),
             SQL::Function.new(:pg_get_expr, :pg_attrdef__adbin, :pg_class__oid).as(:default),
             SQL::BooleanExpression.new(:NOT, :pg_attribute__attnotnull).as(:allow_null),
@@ -767,6 +914,11 @@ module Sequel
         "(#{Array(args).map{|a| Array(a).reverse.join(' ')}.join(', ')})"
       end
 
+      # PostgreSQL can combine multiple alter table ops into a single query.
+      def supports_combining_alter_table_ops?
+        true
+      end
+
       # Handle bigserial type if :serial option is present
       def type_literal_generic_bignum(column)
         column[:serial] ? :bigserial : super
@@ -965,6 +1117,11 @@ module Sequel
         true
       end
 
+      # PostgreSQL supports pattern matching via regular expressions
+      def supports_regexp?
+        true
+      end
+
       # PostgreSQL supports timezones in literal timestamps
       def supports_timestamp_timezones?
         true
