@@ -13,7 +13,7 @@ module Sequel
     # the column method, which makes for a nicer DSL.
     #
     # For more information on Sequel's support for schema modification, see
-    # the {"Migrations and Schema Modification" guide}[link:files/doc/migration_rdoc.html].
+    # the {"Schema Modification" guide}[link:files/doc/schema_modification_rdoc.html].
     class CreateTableGenerator
       # Classes specifying generic types that Sequel will convert to database-specific types.
       GENERIC_TYPES=[String, Integer, Fixnum, Bignum, Float, Numeric, BigDecimal,
@@ -79,10 +79,11 @@ module Sequel
       # The following options are supported:
       #
       # :default :: The default value for the column.
-      # :deferrable :: This ensure Referential Integrity will work even if
-      #                reference table will use for its foreign key a value that does not
-      #                exists(yet) on referenced table. Basically it adds
-      #                DEFERRABLE INITIALLY DEFERRED on key creation.
+      # :deferrable :: For foreign key columns, this ensures referential integrity will work even if
+      #                referencing table uses a foreign key value that does not
+      #                yet exist on referenced table (but will exist before the transaction commits).
+      #                Basically it adds DEFERRABLE INITIALLY DEFERRED on key creation.
+      #                If you use :immediate as the value, uses DEFERRABLE INITIALLY IMMEDIATE.
       # :index :: Create an index on this column.  If given a hash, use the hash as the
       #           options for the index.
       # :key :: For foreign key columns, the column in the associated table
@@ -98,6 +99,8 @@ module Sequel
       #               (:restrict, :cascade, :set_null, :set_default, :no_action).
       # :primary_key :: Make the column as a single primary key column.  This should only
       #                 be used if you have a single, nonautoincrementing primary key column.
+      # :type :: Overrides the type given as the argument.  Generally not used by column
+      #          itself, but can be passed as an option to other methods that call column.
       # :unique :: Mark the column as unique, generally has the same effect as
       #            creating a unique index on the column.
       def column(name, type, opts = {})
@@ -122,6 +125,7 @@ module Sequel
       #   foreign_key(:artist_id) # artist_id INTEGER
       #   foreign_key(:artist_id, :artists) # artist_id INTEGER REFERENCES artists
       #   foreign_key(:artist_id, :artists, :key=>:id) # artist_id INTEGER REFERENCES artists(id)
+      #   foreign_key(:artist_id, :artists, :type=>String) # artist_id varchar(255) REFERENCES artists(id)
       #
       # If you want a foreign key constraint without adding a column (usually because it is a
       # composite foreign key), you can provide an array of columns as the first argument, and
@@ -174,7 +178,7 @@ module Sequel
       # :concurrently :: Create the index concurrently, so it doesn't block
       #                  operations on the table while the index is being
       #                  built.
-      # :op_class :: Use a specific operator class in the index.
+      # :opclass :: Use a specific operator class in the index.
       #
       # Microsoft SQL Server specific options:
       #
@@ -195,6 +199,11 @@ module Sequel
       def method_missing(type, name = nil, opts = {})
         name ? column(name, type, opts) : super
       end
+
+      # This object responds to all methods.
+      def respond_to_missing?(meth, include_private)
+        true
+      end
       
       # Adds an autoincrementing primary key column or a primary key constraint to the DDL.
       # To create a constraint, the first argument should be an array of column symbols
@@ -236,6 +245,8 @@ module Sequel
       # Add a unique constraint on the given columns to the DDL.
       #
       #   unique(:name) # UNIQUE (name)
+      #
+      # Supports the same :deferrable option as #column.
       def unique(columns, opts = {})
         constraints << {:type => :unique, :columns => Array(columns)}.merge(opts)
       end
@@ -267,7 +278,7 @@ module Sequel
     # alter a table's description.
     #
     # For more information on Sequel's support for schema modification, see
-    # the {"Migrations and Schema Modification" guide}[link:files/doc/migration_rdoc.html].
+    # the {"Schema Modification" guide}[link:files/doc/schema_modification_rdoc.html].
     class AlterTableGenerator
       # An array of DDL operations to perform
       attr_reader :operations
@@ -291,7 +302,7 @@ module Sequel
       # Add a constraint with the given name and args to the DDL for the table.
       # See CreateTableGenerator#constraint.
       #
-      #   add_constraint(:valid_name, :name.like('A%'))
+      #   add_constraint(:valid_name, Sequel.like(:name, 'A%'))
       #   # ADD CONSTRAINT valid_name CHECK (name LIKE 'A%')
       def add_constraint(name, *args, &block)
         @operations << {:op => :add_constraint, :name => name, :type => :check, :check => block || args}
@@ -301,6 +312,8 @@ module Sequel
       #
       #   add_unique_constraint(:name) # ADD UNIQUE (name)
       #   add_unique_constraint(:name, :name=>:unique_name) # ADD CONSTRAINT unique_name UNIQUE (name)
+      #
+      # Supports the same :deferrable option as CreateTableGenerator#column.
       def add_unique_constraint(columns, opts = {})
         @operations << {:op => :add_constraint, :type => :unique, :columns => Array(columns)}.merge(opts)
       end
@@ -427,13 +440,20 @@ module Sequel
         @operations << {:op => :set_column_type, :name => name, :type => type}.merge(opts)
       end
       
-      # Modify a column's NOT NULL constraint.
+      # Set a given column as allowing NULL values.
       #
-      #   set_column_allow_null(:artist_name, false) # ALTER COLUMN artist_name SET NOT NULL
-      def set_column_allow_null(name, allow_null)
+      #   set_column_allow_null(:artist_name) # ALTER COLUMN artist_name DROP NOT NULL
+      def set_column_allow_null(name, allow_null=true)
         @operations << {:op => :set_column_null, :name => name, :null => allow_null}
       end
 
+      # Set a given column as not allowing NULL values.
+      #
+      #   set_column_not_null(:artist_name) # ALTER COLUMN artist_name SET NOT NULL
+      def set_column_not_null(name)
+        set_column_allow_null(name, false)
+      end
+
       private
 
       # Add a composite primary key constraint
