@@ -22,10 +22,18 @@ describe "A new Database" do
     Sequel::Database.new(1 => 2, :logger => 4, :loggers => 3).loggers.should == [4,3]
     Sequel::Database.new(1 => 2, :logger => [4], :loggers => [3]).loggers.should == [4,3]
   end
+
+  specify "should handle the default string column size" do
+    @db.default_string_column_size.should == 255
+    db = Sequel::Database.new(:default_string_column_size=>50)
+    db.default_string_column_size.should == 50
+    db.default_string_column_size = 2
+    db.default_string_column_size.should == 2
+  end
   
   specify "should set the sql_log_level from opts[:sql_log_level]" do
-    db = Sequel::Database.new(1 => 2, :sql_log_level=>:debug).sql_log_level.should == :debug
-    db = Sequel::Database.new(1 => 2, :sql_log_level=>'debug').sql_log_level.should == :debug
+    Sequel::Database.new(1 => 2, :sql_log_level=>:debug).sql_log_level.should == :debug
+    Sequel::Database.new(1 => 2, :sql_log_level=>'debug').sql_log_level.should == :debug
   end
   
   specify "should create a connection pool" do
@@ -37,7 +45,8 @@ describe "A new Database" do
   
   specify "should pass the supplied block to the connection pool" do
     cc = nil
-    d = Sequel::Database.new {1234}
+    d = Sequel::Database.new
+    d.meta_def(:connect){|c| 1234}
     d.synchronize {|c| cc = c}
     cc.should == 1234
   end
@@ -369,7 +378,7 @@ describe "Database#dataset" do
   end
   
   specify "should provide a filtered #from dataset if a block is given" do
-    d = @db.from(:mau) {:x.sql_number > 100}
+    d = @db.from(:mau){x.sql_number > 100}
     d.should be_a_kind_of(Sequel::Dataset)
     d.sql.should == 'SELECT * FROM mau WHERE (x > 100)'
   end
@@ -538,7 +547,8 @@ end
 
 describe "Database#synchronize" do
   before do
-    @db = Sequel::Database.new(:max_connections => 1){12345}
+    @db = Sequel::Database.new(:max_connections => 1)
+    @db.meta_def(:connect){|c| 12345}
   end
   
   specify "should wrap the supplied block in pool.hold" do
@@ -561,7 +571,9 @@ end
 
 describe "Database#test_connection" do
   before do
-    @db = Sequel::Database.new{@test = rand(100)}
+    @db = Sequel::Database.new
+    pr = proc{@test = rand(100)}
+    @db.meta_def(:connect){|c| pr.call}
   end
   
   specify "should attempt to get a connection" do
@@ -617,6 +629,18 @@ shared_examples_for "Database#transaction" do
                        'BEGIN', 'SET TRANSACTION ISOLATION LEVEL SERIALIZABLE', 'DROP TABLE serializable', 'COMMIT']
   end
   
+  specify "should support :disconnect=>:retry option for automatically retrying on disconnect" do
+    a = []
+    @db.transaction(:disconnect=>:retry){a << 1; raise Sequel::DatabaseDisconnectError if a.length < 2}
+    @db.sqls.should == ['BEGIN', 'ROLLBACK', 'BEGIN', 'COMMIT']
+    a.should == [1, 1]
+  end
+  
+  specify "should raise an error if attempting to use :disconnect=>:retry inside another transaction" do
+    proc{@db.transaction{@db.transaction(:disconnect=>:retry){}}}.should raise_error(Sequel::Error)
+    @db.sqls.should == ['BEGIN', 'ROLLBACK']
+  end
+  
   specify "should handle returning inside of the block by committing" do
     def @db.ret_commit
       transaction do
@@ -659,7 +683,7 @@ shared_examples_for "Database#transaction" do
     end
     e.should_not be_nil
     e.wrapped_exception.should be_a_kind_of(ec)
-    @db.sqls.should == ['BEGIN', 'DROP TABLE test;']
+    @db.sqls.should == ['BEGIN', 'DROP TABLE test;', 'ROLLBACK']
   end
   
   specify "should handle errors when sending ROLLBACK" do
@@ -1246,7 +1270,8 @@ end
 describe "A single threaded database" do
   before do
     conn = 1234567
-    @db = Sequel::Database.new(:single_threaded => true) do
+    @db = Sequel::Database.new(:single_threaded => true)
+    @db.meta_def(:connect) do |c|
       conn += 1
     end
   end
@@ -1360,7 +1385,7 @@ describe "Database#fetch" do
     ds.select_sql.should == 'select * from xyz'
     ds.sql.should == 'select * from xyz'
     
-    ds.filter!(:price.sql_number < 100)
+    ds.filter!{price.sql_number < 100}
     ds.select_sql.should == 'select * from xyz'
     ds.sql.should == 'select * from xyz'
   end
@@ -1409,7 +1434,7 @@ describe "Database#get" do
     @db.get(1).should == 1
     @db.sqls.should == ['SELECT 1 LIMIT 1']
     
-    @db.get(:version.sql_function)
+    @db.get(Sequel.function(:version))
     @db.sqls.should == ['SELECT version() LIMIT 1']
   end
 
@@ -1726,16 +1751,17 @@ describe "Database#typecast_value" do
   end
 
   specify "should typecast datetime values to Sequel.datetime_class with correct timezone handling" do
-    t = Time.utc(2011, 1, 2, 3, 4, 5) # UTC Time
-    t2 = Time.mktime(2011, 1, 2, 3, 4, 5) # Local Time
-    t3 = Time.utc(2011, 1, 2, 3, 4, 5) - (t - t2) # Local Time in UTC Time
-    t4 = Time.mktime(2011, 1, 2, 3, 4, 5) + (t - t2) # UTC Time in Local Time
-    dt = DateTime.civil(2011, 1, 2, 3, 4, 5)
+    t = Time.utc(2011, 1, 2, 3, 4, 5, 500000) # UTC Time
+    t2 = Time.mktime(2011, 1, 2, 3, 4, 5, 500000) # Local Time
+    t3 = Time.utc(2011, 1, 2, 3, 4, 5, 500000) - (t - t2) # Local Time in UTC Time
+    t4 = Time.mktime(2011, 1, 2, 3, 4, 5, 500000) + (t - t2) # UTC Time in Local Time
+    secs = defined?(Rational) ? Rational(11, 2) : 5.5
     r1 = defined?(Rational) ? Rational(t2.utc_offset, 86400) : t2.utc_offset/86400.0
     r2 = defined?(Rational) ? Rational((t - t2).to_i, 86400) : (t - t2).to_i/86400.0
-    dt2 = DateTime.civil(2011, 1, 2, 3, 4, 5, r1)
-    dt3 = DateTime.civil(2011, 1, 2, 3, 4, 5) - r2
-    dt4 = DateTime.civil(2011, 1, 2, 3, 4, 5, r1) + r2
+    dt = DateTime.civil(2011, 1, 2, 3, 4, secs)
+    dt2 = DateTime.civil(2011, 1, 2, 3, 4, secs, r1)
+    dt3 = DateTime.civil(2011, 1, 2, 3, 4, secs) - r2
+    dt4 = DateTime.civil(2011, 1, 2, 3, 4, secs, r1) + r2
 
     t.should == t4
     t2.should == t3
@@ -1751,25 +1777,26 @@ describe "Database#typecast_value" do
         v.offset.should == o.offset
       end
     end
+    @db.extend_datasets(Module.new{def supports_timestamp_timezones?; true; end})
     begin
       @db.typecast_value(:datetime, dt).should == t
       @db.typecast_value(:datetime, dt2).should == t2
       @db.typecast_value(:datetime, t).should == t
       @db.typecast_value(:datetime, t2).should == t2
-      @db.typecast_value(:datetime, dt.to_s).should == t
-      @db.typecast_value(:datetime, dt.strftime('%F %T')).should == t2
+      @db.typecast_value(:datetime, @db.literal(dt)[1...-1]).should == t
+      @db.typecast_value(:datetime, dt.strftime('%F %T.%N')).should == t2
       @db.typecast_value(:datetime, Date.civil(2011, 1, 2)).should == Time.mktime(2011, 1, 2, 0, 0, 0)
-      @db.typecast_value(:datetime, :year=>dt.year, :month=>dt.month, :day=>dt.day, :hour=>dt.hour, :minute=>dt.min, :second=>dt.sec).should == t2
+      @db.typecast_value(:datetime, :year=>dt.year, :month=>dt.month, :day=>dt.day, :hour=>dt.hour, :minute=>dt.min, :second=>dt.sec, :nanos=>500000000).should == t2
 
       Sequel.datetime_class = DateTime
       @db.typecast_value(:datetime, dt).should == dt
       @db.typecast_value(:datetime, dt2).should == dt2
       @db.typecast_value(:datetime, t).should == dt
       @db.typecast_value(:datetime, t2).should == dt2
-      @db.typecast_value(:datetime, dt.to_s).should == dt
-      @db.typecast_value(:datetime, dt.strftime('%F %T')).should == dt
+      @db.typecast_value(:datetime, @db.literal(dt)[1...-1]).should == dt
+      @db.typecast_value(:datetime, dt.strftime('%F %T.%N')).should == dt
       @db.typecast_value(:datetime, Date.civil(2011, 1, 2)).should == DateTime.civil(2011, 1, 2, 0, 0, 0)
-      @db.typecast_value(:datetime, :year=>dt.year, :month=>dt.month, :day=>dt.day, :hour=>dt.hour, :minute=>dt.min, :second=>dt.sec).should == dt
+      @db.typecast_value(:datetime, :year=>dt.year, :month=>dt.month, :day=>dt.day, :hour=>dt.hour, :minute=>dt.min, :second=>dt.sec, :nanos=>500000000).should == dt
 
       Sequel.application_timezone = :utc
       Sequel.typecast_timezone = :local
@@ -1778,20 +1805,20 @@ describe "Database#typecast_value" do
       check[dt2, t3]
       check[t, t]
       check[t2, t3]
-      check[dt.to_s, t]
-      check[dt.strftime('%F %T'), t3]
+      check[@db.literal(dt)[1...-1], t]
+      check[dt.strftime('%F %T.%N'), t3]
       check[Date.civil(2011, 1, 2), Time.utc(2011, 1, 2, 0, 0, 0)]
-      check[{:year=>dt.year, :month=>dt.month, :day=>dt.day, :hour=>dt.hour, :minute=>dt.min, :second=>dt.sec}, t3]
+      check[{:year=>dt.year, :month=>dt.month, :day=>dt.day, :hour=>dt.hour, :minute=>dt.min, :second=>dt.sec, :nanos=>500000000}, t3]
 
       Sequel.datetime_class = DateTime
       check[dt, dt]
       check[dt2, dt3]
       check[t, dt]
       check[t2, dt3]
-      check[dt.to_s, dt]
-      check[dt.strftime('%F %T'), dt3]
+      check[@db.literal(dt)[1...-1], dt]
+      check[dt.strftime('%F %T.%N'), dt3]
       check[Date.civil(2011, 1, 2), DateTime.civil(2011, 1, 2, 0, 0, 0)]
-      check[{:year=>dt.year, :month=>dt.month, :day=>dt.day, :hour=>dt.hour, :minute=>dt.min, :second=>dt.sec}, dt3]
+      check[{:year=>dt.year, :month=>dt.month, :day=>dt.day, :hour=>dt.hour, :minute=>dt.min, :second=>dt.sec, :nanos=>500000000}, dt3]
 
       Sequel.typecast_timezone = :utc
       Sequel.datetime_class = Time
@@ -1799,20 +1826,20 @@ describe "Database#typecast_value" do
       check[dt2, t3]
       check[t, t]
       check[t2, t3]
-      check[dt.to_s, t]
-      check[dt.strftime('%F %T'), t]
+      check[@db.literal(dt)[1...-1], t]
+      check[dt.strftime('%F %T.%N'), t]
       check[Date.civil(2011, 1, 2), Time.utc(2011, 1, 2, 0, 0, 0)]
-      check[{:year=>dt.year, :month=>dt.month, :day=>dt.day, :hour=>dt.hour, :minute=>dt.min, :second=>dt.sec}, t]
+      check[{:year=>dt.year, :month=>dt.month, :day=>dt.day, :hour=>dt.hour, :minute=>dt.min, :second=>dt.sec, :nanos=>500000000}, t]
 
       Sequel.datetime_class = DateTime
       check[dt, dt]
       check[dt2, dt3]
       check[t, dt]
       check[t2, dt3]
-      check[dt.to_s, dt]
-      check[dt.strftime('%F %T'), dt]
+      check[@db.literal(dt)[1...-1], dt]
+      check[dt.strftime('%F %T.%N'), dt]
       check[Date.civil(2011, 1, 2), DateTime.civil(2011, 1, 2, 0, 0, 0)]
-      check[{:year=>dt.year, :month=>dt.month, :day=>dt.day, :hour=>dt.hour, :minute=>dt.min, :second=>dt.sec}, dt]
+      check[{:year=>dt.year, :month=>dt.month, :day=>dt.day, :hour=>dt.hour, :minute=>dt.min, :second=>dt.sec, :nanos=>500000000}, dt]
 
       Sequel.application_timezone = :local
       Sequel.datetime_class = Time
@@ -1820,20 +1847,20 @@ describe "Database#typecast_value" do
       check[dt2, t2]
       check[t, t4]
       check[t2, t2]
-      check[dt.to_s, t4]
-      check[dt.strftime('%F %T'), t4]
+      check[@db.literal(dt)[1...-1], t4]
+      check[dt.strftime('%F %T.%N'), t4]
       check[Date.civil(2011, 1, 2), Time.local(2011, 1, 2, 0, 0, 0)]
-      check[{:year=>dt.year, :month=>dt.month, :day=>dt.day, :hour=>dt.hour, :minute=>dt.min, :second=>dt.sec}, t4]
+      check[{:year=>dt.year, :month=>dt.month, :day=>dt.day, :hour=>dt.hour, :minute=>dt.min, :second=>dt.sec, :nanos=>500000000}, t4]
 
       Sequel.datetime_class = DateTime
       check[dt, dt4]
       check[dt2, dt2]
       check[t, dt4]
       check[t2, dt2]
-      check[dt.to_s, dt4]
-      check[dt.strftime('%F %T'), dt4]
+      check[@db.literal(dt)[1...-1], dt4]
+      check[dt.strftime('%F %T.%N'), dt4]
       check[Date.civil(2011, 1, 2), DateTime.civil(2011, 1, 2, 0, 0, 0, r1)]
-      check[{:year=>dt.year, :month=>dt.month, :day=>dt.day, :hour=>dt.hour, :minute=>dt.min, :second=>dt.sec}, dt4]
+      check[{:year=>dt.year, :month=>dt.month, :day=>dt.day, :hour=>dt.hour, :minute=>dt.min, :second=>dt.sec, :nanos=>500000000}, dt4]
 
       Sequel.typecast_timezone = :local
       Sequel.datetime_class = Time
@@ -1841,20 +1868,20 @@ describe "Database#typecast_value" do
       check[dt2, t2]
       check[t, t4]
       check[t2, t2]
-      check[dt.to_s, t4]
-      check[dt.strftime('%F %T'), t2]
+      check[@db.literal(dt)[1...-1], t4]
+      check[dt.strftime('%F %T.%N'), t2]
       check[Date.civil(2011, 1, 2), Time.local(2011, 1, 2, 0, 0, 0)]
-      check[{:year=>dt.year, :month=>dt.month, :day=>dt.day, :hour=>dt.hour, :minute=>dt.min, :second=>dt.sec}, t2]
+      check[{:year=>dt.year, :month=>dt.month, :day=>dt.day, :hour=>dt.hour, :minute=>dt.min, :second=>dt.sec, :nanos=>500000000}, t2]
 
       Sequel.datetime_class = DateTime
       check[dt, dt4]
       check[dt2, dt2]
       check[t, dt4]
       check[t2, dt2]
-      check[dt.to_s, dt4]
-      check[dt.strftime('%F %T'), dt2]
+      check[@db.literal(dt)[1...-1], dt4]
+      check[dt.strftime('%F %T.%N'), dt2]
       check[Date.civil(2011, 1, 2), DateTime.civil(2011, 1, 2, 0, 0, 0, r1)]
-      check[{:year=>dt.year, :month=>dt.month, :day=>dt.day, :hour=>dt.hour, :minute=>dt.min, :second=>dt.sec}, dt2]
+      check[{:year=>dt.year, :month=>dt.month, :day=>dt.day, :hour=>dt.hour, :minute=>dt.min, :second=>dt.sec, :nanos=>500000000}, dt2]
 
     ensure
       Sequel.default_timezone = nil
@@ -1914,7 +1941,7 @@ describe "Database#typecast_value" do
   end
 
   specify "should typecast string values to String" do
-    [1.0, '1.0', '1.0'.to_sequel_blob].each do |i|
+    [1.0, '1.0', Sequel.blob('1.0')].each do |i|
       v = @db.typecast_value(:string, i)
       v.should be_an_instance_of(String)
       v.should == "1.0"
@@ -2063,15 +2090,18 @@ end
 describe "Database#column_schema_to_ruby_default" do
   specify "should handle converting many default formats" do
     db = Sequel::Database.new
-    m = db.method(:column_schema_to_ruby_default)
-    p = lambda{|d,t| m.call(d,t)}
+    p = lambda{|d,t| db.send(:column_schema_to_ruby_default, d, t)}
     p[nil, :integer].should be_nil
+    p[1, :integer].should == 1
     p['1', :integer].should == 1
     p['-1', :integer].should == -1
+    p[1.0, :float].should == 1.0
     p['1.0', :float].should == 1.0
     p['-1.0', :float].should == -1.0
     p['1.0', :decimal].should == BigDecimal.new('1.0')
     p['-1.0', :decimal].should == BigDecimal.new('-1.0')
+    p[true, :boolean].should == true
+    p[false, :boolean].should == false
     p['1', :boolean].should == true
     p['0', :boolean].should == false
     p['true', :boolean].should == true
@@ -2079,33 +2109,40 @@ describe "Database#column_schema_to_ruby_default" do
     p["'t'", :boolean].should == true
     p["'f'", :boolean].should == false
     p["'a'", :string].should == 'a'
-    p["'a'", :blob].should == 'a'.to_sequel_blob
+    p["'a'", :blob].should == Sequel.blob('a')
     p["'a'", :blob].should be_a_kind_of(Sequel::SQL::Blob)
     p["''", :string].should == ''
     p["'\\a''b'", :string].should == "\\a'b"
     p["'NULL'", :string].should == "NULL"
+    p[Date.today, :date].should == Date.today
     p["'2009-10-29'", :date].should == Date.new(2009,10,29)
-    p["CURRENT_TIMESTAMP", :date].should be_nil
-    p["today()", :date].should be_nil
+    p["CURRENT_TIMESTAMP", :date].should == Sequel::CURRENT_DATE
+    p["CURRENT_DATE", :date].should == Sequel::CURRENT_DATE
+    p["now()", :date].should == Sequel::CURRENT_DATE
+    p["getdate()", :date].should == Sequel::CURRENT_DATE
+    p["CURRENT_TIMESTAMP", :datetime].should == Sequel::CURRENT_TIMESTAMP
+    p["CURRENT_DATE", :datetime].should == Sequel::CURRENT_TIMESTAMP
+    p["now()", :datetime].should == Sequel::CURRENT_TIMESTAMP
+    p["getdate()", :datetime].should == Sequel::CURRENT_TIMESTAMP
     p["'2009-10-29T10:20:30-07:00'", :datetime].should == DateTime.parse('2009-10-29T10:20:30-07:00')
     p["'2009-10-29 10:20:30'", :datetime].should == DateTime.parse('2009-10-29 10:20:30')
     p["'10:20:30'", :time].should == Time.parse('10:20:30')
     p["NaN", :float].should be_nil
 
-    db.meta_def(:database_type){:postgres}
+    db = Sequel.mock(:host=>'postgres')
     p["''::text", :string].should == ""
     p["'\\a''b'::character varying", :string].should == "\\a'b"
     p["'a'::bpchar", :string].should == "a"
     p["(-1)", :integer].should == -1
     p["(-1.0)", :float].should == -1.0
     p['(-1.0)', :decimal].should == BigDecimal.new('-1.0')
-    p["'a'::bytea", :blob].should == 'a'.to_sequel_blob
+    p["'a'::bytea", :blob].should == Sequel.blob('a')
     p["'a'::bytea", :blob].should be_a_kind_of(Sequel::SQL::Blob)
     p["'2009-10-29'::date", :date].should == Date.new(2009,10,29)
     p["'2009-10-29 10:20:30.241343'::timestamp without time zone", :datetime].should == DateTime.parse('2009-10-29 10:20:30.241343')
     p["'10:20:30'::time without time zone", :time].should == Time.parse('10:20:30')
 
-    db.meta_def(:database_type){:mysql}
+    db = Sequel.mock(:host=>'mysql')
     p["\\a'b", :string].should == "\\a'b"
     p["a", :string].should == "a"
     p["NULL", :string].should == "NULL"
@@ -2114,11 +2151,10 @@ describe "Database#column_schema_to_ruby_default" do
     p["2009-10-29", :date].should == Date.new(2009,10,29)
     p["2009-10-29 10:20:30", :datetime].should == DateTime.parse('2009-10-29 10:20:30')
     p["10:20:30", :time].should == Time.parse('10:20:30')
-    p["CURRENT_DATE", :date].should be_nil
-    p["CURRENT_TIMESTAMP", :datetime].should be_nil
     p["a", :enum].should == "a"
+    p["a,b", :set].should == "a,b"
     
-    db.meta_def(:database_type){:mssql}
+    db = Sequel.mock(:host=>'mssql')
     p["(N'a')", :string].should == "a"
     p["((-12))", :integer].should == -12
     p["((12.1))", :float].should == 12.1
